# js进阶知识 (下)

## (一) 面向过程编程和面向对象编程

**面向过程 **Procedure-Oriented Programming(简称POP), 就是分析出解决问题所需要的步骤，然后用函数把这些步骤一步一步实现，使用的时候一个一个依次调用就可以了。

**面向对象** Object-Oriented Programming(简称OOP), 是把构成问题事务分解成各个对象，建立对象的目的不是为了完成一个步骤，而是为了描叙某个事物在整个解决问题的步骤中的行为。 

#### 例1. 面向过程

获取列表数据, 渲染页面

```javascript
<ul></ul>

<script>
    var xhr = new XMLHttpRequest();
    xhr.open('get', 'http://81.71.65.4:3003/category/all');
    xhr.send();
    xhr.onreadystatechange = () => {
        if (xhr.readyState === 4 && xhr.status === 200) {
            var res = JSON.parse(xhr.responseText);
            var list = res.list;
            var htmlStr = '';
            list.forEach(item => {
                htmlStr += `<li>${item.name}</li> `
            })
            document.querySelector('ul').innerHTML = htmlStr;
        }
    }
</script>
```

优点: 逻辑清晰, 编写方便(这也是大家喜欢的原因)

缺点1: 代码难以维护, 如果需要的做的事情更多, 那么代码就会更多, 出错了就不好维护

缺点2: 如果后面需要重新获取这个接口数据(比如分页), 就需要重新写一遍, 代码无法复用, 容易产生冗余

#### 例2. 面向对象编程

```JavaScript
<body>
    <p>
        <button onclick="obj.changePage(1)">1</button>
        <button onclick="obj.changePage(2)">2</button>
        <button onclick="obj.changePage(3)">3</button>
    </p>
    <ul></ul>

    <script>
        var obj = {
            init() {
                this.getList();
            },

            // 切换分页
            changePage(num) {
                this.data.currPage = num;
                this.getList();
            },

            data: {
                currPage: 1,
                list: [],
            },
 

            getList() {
                var xhr = new XMLHttpRequest();
                let currPage = this.data.currPage;
                xhr.open('get', `./json/list${currPage}.json`);
                xhr.send();
                xhr.onreadystatechange = () => {
                    if (xhr.readyState === 4 && xhr.status === 200) {
                        var res = JSON.parse(xhr.responseText);
                        this.data.list = res.list;
                        this.render();
                    }
                }
            },

            render() {
                let list = this.data.list;
                var htmlStr = '';
                list.forEach(item => {
                    htmlStr += `<li>${item.name}</li> `
                })
                document.querySelector('ul').innerHTML = htmlStr;
            }
        }
        obj.init();
    </script>
</body>
```

优点: 职责分明, 代码可以很好复用

缺点: 对新手不友好, 各个方法相互之间调来调去, 容易混乱

总结: 其实面向对象编程中也同时在使用面向过程的思想, 比如每个方法的实现过程就是面向过程

## (二) js创建对象

#### (1) 使用工厂模式创建对象

```JavaScript
function createPerson(name,age,job) {
    var o = new Object();
    o.name = name;
    o.age = age;
    o.job = job;
    return o;
}
var p1 = createPerson('张三',29,'测试工程师');
var p2 = createPerson('李四',28,'前端工程师');
```

#### (2) 使用构造函数模式创建对象

```JavaScript
function Person(name, age, job) { 
  this.name = name;
  this.age = age;
  this.job = job; 
  this.sayName = function() {
    console.log(this.name);
  }
}
 
var p1 = new Person("张三", 29, "测试工程师");
var p2 = new Person("李四", 28, "前端工程师");
p1.sayName();
p2.sayName();
```

**js中的new ()到底做了些什么**

使用new来创建对象时, 系统"偷偷的"做了以下动作:

(1) 创建一个新对象；
(2) 将构造函数的作用域赋给新对象（因此 this 就指向了这个新对象） ；
(3) 执行构造函数中的代码（为这个新对象添加属性和方法） ；
(4) 返回新对象。

#### (3) 使用原型模式创建对象

下面的例子讲

```JavaScript

```

#### 

## (三) 面向对象编程三大特性

https://www.cnblogs.com/zhengyufeng/p/10992972.html

https://blog.csdn.net/WXS153322/article/details/124053548



![img](img/v2-f5785bde2f3a990acf41ea4dbe2b1563_720w.81a892fc.jpg)

### (1) 封装

我们平时所用的方法和类都是一种封装，当我们在项目开发中，遇到一段功能的代码在好多地方重复使用的时候，我们可以把他单独封装成一个功能的方法，这样在我们需要使用的地方直接调用就可以了。

```JavaScript
var obj = {
    init() {},
    getList() {},
    render() {}
}
obj.init();
```

### (2) 继承

#### 01 es6的class(类)

在es6中，提供了新的方式去书写面向对象编程，这种方式更加接近面向对象的书写方式，可以通过 class 关键字定义类。它的本质是函数（function）,但本身只是一种语法糖

```javascript
// 声明一个类
class Cat {
  // 声明静态属性
  static type = '猫科动物';

  // 构造器(构造函数), 名字是固定的
  constructor(name, age) {
    // 构造器里所做的事情, 跟es5的构造函数做的事情是一样
    this.name = name;
    this.age = age;
  }

  // 类的方法
  say(){
    console.log(this.name);
  }

  // 静态方法
  static climbTree() {
      console.log('猫会爬树');
  } 
}


const cat = new Cat('小花', 2);
console.log(cat);
cat.say();

// 静态属性和静态方法由类本身直接调用
console.log('静态属性',Cat.type);
Cat.climbTree();
```

#### 02 es6的class继承

es5的继承方式请看后面的例子

```JavaScript
class Father {
  static type = "父类";
  constructor(name) {
    this.name = name;
  }
  static say() {
    console.log("哈哈哈");
  }

  getName() {
    console.log(this.name);
  }
}

class Son extends Father {
  constructor(name, age) {
    // super用来调用父类构造器
    super(name);
    this.age = age;
  }
}

const son = new Son("小三", 3);
console.log(son);
son.getName();

// 静态属性和方法不继承
console.log('type', son.type);
```

特别提醒：继承会继承父类的实例属性和实例方法，并不会继承静态属性和静态方法，并且静态方法只能通过类名去调用。

### (3) 多态

多态的具体表现为方法重载和方法重写：

方法重载：重载是指不同的函数使用相同的函数名，但是函数的参数个数或类型不同。调用的时候根据函数的参数来区别不同的函数

方法重写：重写（也叫覆盖）是指在派生类中重新对基类中的虚函数（注意是虚函数）重新实现。即函数名和参数都一样，只是函数的实现体不一样

#### 01 方法重写

```JavaScript
class Father {
  constructor(name) {
    this.name = name;
  }

  work() {
    console.log('我的工作是累死累活, 赚钱养家')
  }
}


class Son extends Father {
  constructor(name) {
    super(name);
  }
  // 重写方法
  work() {
    console.log('我的工作是好好学习, 天天向上');
  }  
}

// 父亲的实例
const zhangSan = new Father('张三');
zhangSan.work();
// 儿子的实例
const xiaoSan = new Son('小三');
xiaoSan.work();
```



#### 02 方法重载

js无法使用重载, 因为若多个函数使用了相同的函数名, 后面的会覆盖前面的, 下面是java的重载

```JavaScript
public class Human {
 String hName;
 //无参构造方法
 Human(){
  System.out.println("创建人类对象，默认姓名为路人甲！");
  hName = "路人甲";
 }
 //带参数构造方法
 Human(String name){
  System.out.println("创建人类对象，由调用者自行指定姓名！");
  hName=name; 
 }
 //吃饭，午餐
 void eat() {
  System.out.println(hName+"说：我使用筷子吃饭！");
 }
 //吃饭，可以指定工具
 void eat(String tool) {
  System.out.println(hName+"说：我使用"+tool+"吃饭！");
 }
}
 
 
public class HumanTest {
 public static void main(String[] args) {
  //使用无参构造方法创建人类对象，并调用无参eat()方法
  Human h1 = new Human();
  h1.eat();
  //使用了另外一条生产线，并指定吃饭工具
  Human h2 = new Human("Jack");
  h2.eat("刀和叉");
 }
}
```





## (四) 原型和原型链

### (1) 原型(原型对象)

1. 每一个构造函数都有一个prototype属性, 指向了构造函数的原型对象
2. 实例继承原型对象所有的属性和方法 
3. 实例有一个内部指针`__proto__`指向了它的构造函数的原型对象
4. 原型对象有一个constructor指针指向了构造函数

#### 例1. 构造函数&原型对象&实例

* 在控制台通过dir(Cat)来查看Cat的属性和方法

* instanceof可以判断实例和构造函数的关系

```JavaScript
<script>
    // 声明一个构造函数
    function Cat(name, age) {
        this.name = name;
        this.age = age;
    }
    var cat = new Cat('小花',10);
    // console.log('构造函数',Cat);
    console.log('实例', cat);
    console.log('原型对象', Cat.prototype);
    console.log(Cat.prototype === cat.__proto__);  // true
    // 判断实例和构造函数的关系
    console.log(cat instanceof Cat); // true
</script>
```

三者的关系图

![image-20220930154736246](../docs/mg/image-20220930154736246.png)



#### 例2. 给原型对象添加实例和方法

由构造函数创建出来的实例会继承原型对象上所有的属性和方法

```JavaScript
<script>
    // 声明一个构造函数
    function Cat(name, age) {
        this.name = name;
        this.age = age;
    }

    Cat.prototype.type = '猫科动物';
    Cat.prototype.say = function () { 
        console.log(this.name);
    }

    var cat = new Cat("小花", 10);
    console.log(cat.type);
    cat.say();
</script>
```

#### 例3. 区分属性和方法的归属

hasOwnProperty是否是实例的属性, own的意思是自己的,本人的

```JavaScript
<script>
    // 声明一个构造函数
    function Cat(name, age) {
        this.name = name;
        this.age = age;
    }

    Cat.prototype.type = '猫科动物';
    Cat.prototype.say = function () { 
        console.log(this.name);
    }

    var cat = new Cat("小花", 10); 

    for(var key in cat) { 
        if (cat.hasOwnProperty(key)) {
            console.log('实例的属性',key);
        } else {
            console.log('继承来的属性',key);
        }
    } 
</script>
```

ps: 递归深拷贝一般只是拷贝对象本身的属性, 继承来的属性不拷贝, 所以应该判断一下是不是对象自己的属性

```javascript
<script>
    function deepCopy(o) {
        var newO = Array.isArray(o) ? [] : {};
        for (key in o) {
            // 只拷贝属性自己的属性
            if (o.hasOwnProperty(key)) {
                // 判断是不是对象
                if (typeof o[key] === 'object' && o[key]) {
                    // 拷贝下一层
                    newO[key] = deepCopy(o[key]);
                } else {
                    newO[key] = o[key];
                }
            }
        }
        return newO;
    }


    console.log('递归调用深拷贝')
    var obj = {
        a: 2,
        b: 3,
        cat: {
            name: "小花",
            addr: {
                msg: '美国'
            }
        }
    };
    var obj2 = deepCopy(obj);
    console.log(obj2.cat.addr === obj.cat.addr);
</script>
```



### (2) 原型链

1. 每一个构造函数都有一个prototype属性指向了原型对象, 同时它的实例都有一个内部指针指向了原型对象,
2. 如果**原型对象同时是另外一个类型的实例**, 那么原型对象内部也会有一个内部指针指向了另外一个原型对象
3. 假如另外一原型对象又是另外一个类型的实例, 上述关系依然成立, 如此层层递进就形成了实例与原型对象的链接, 称为原型链
4. 原型链的尽头是Object的原型对象, 印证了"一切皆对象"这句话

#### 例1. 原型链

**第1句话:** 构造函数Son.prototype指向了原型对象, 它的实例son有一个内部指针指向了原型对象, 也就是`son.__proto__` === Son.prototype

**第2句话:** Son原型对象是另外一个类(Father)的一个实例

`var zhangSan = new Father();`

`Son.prototype = zhangSan, 

那么`Son.prototype.__proto__ === Father.prototype`, `xiaoSan.__proto__.__proto__ === Father.prototype`

**第3句话:**(略) 

**第4句话:** `son.__proto__.__proto__.__proto__ === Object.prototype`

```JavaScript
<script>
    function GrandFater() {
        this.grandFaterName = '老三';
        this.grandFaterAge = 100;
    }
    var laoSan = new GrandFater();
    Father.prototype = laoSan;

    function Father() {
        this.fatherName = '张三';
        this.fatherAge = 50;
    }

    function Son() {
        this.sonName = '小三';
        this.sonAge = 3;
    }
    var zhangSan = new Father();
    Son.prototype = zhangSan;
    var xiaoSan = new Son();
    console.log(xiaoSan.__proto__ === Son.prototype);  
    console.log(Son.prototype === zhangSan);
    console.log(xiaoSan.__proto__ === zhangSan);
    console.log(zhangSan.__proto__ === Father.prototype);
    console.log(xiaoSan.__proto__.__proto__ === Father.prototype);
    // .... 

    console.log('原型链的尽头是:Object.prototype')
    console.log(xiaoSan.__proto__.__proto__.__proto__.__proto__ === Object.prototype);

</script>
```

#### 例2. 一切皆对象

```javascript
    var date = new Date();
    console.log(date.__proto__.__proto__ === Object.prototype);

    var arr = new Array();
    arr[0] = 'hahahah';
    console.log(arr.__proto__.__proto__ === Object.prototype);
```

## (五) es5继承 

#### **(1) 原型链继承**

```js
<script>
    // 父类
    function Father() {
        this.type = '人类';
        this.hobby = ['篮球', '羽毛球'];
    }
    Father.prototype.getType = function () {
        console.log(this.type);
    }
    var father = new Father();
    Son.prototype = father;
    // 子类
    function Son() {}


    var son1 = new Son();    
    var son2 = new Son();   
    // hobby的指针是指向同一个对象
    son2.hobby.push('足球');
    console.log(son1.hobby); 
</script>
```

原型继承的缺点：

- 父类的引用类型属性会被所有子类实例共享，任何一个子类实例修改了父类的引用类型属性，其他子类实例都会受到影响
- 创建子类实例的时候，不能向父类传参

#### **(2) 借用构造函数继承**

```js
<script>
    // 父类
    function Father(name) {
        // 使用call, 所以this指向了son
        this.type = '人类';
        this.name = name;
        this.say = function () {
            console.log('哈哈哈')
        }
        this.hobby = ['篮球'];
    }
    Father.prototype.getType = function () {
        console.log(this.type);
    }

    // 子类
    function Son(name, age) {
        Father.call(this, name);
        this.age = age;
    }

    var son = new Son('小明', 10);
    console.log(son);

    var son2 = new Son('小红', 12);
    son.hobby.push('足球');
    console.log(son2.hobby);

    // 缺点
    // son2.getType(); // 报错, 因为没有 new Father, 所以无法继承Father原型对象上的方法
</script>
```

优点：

1. 避免了引用类型属性被所有实例共享
2. 可以向父类传参

缺点：

1. 方法必须定义在构造函数中
2. 每创建一个实例都会创建一遍方法

#### **(3) 组合式继承（原型继承和借用构造函数继承的组合）**

```js
<script>
    // 父类
    function Father(name) {
        // 使用call, 所以this指向了son
        this.type = '人类';
        this.name = name;
        this.say = function () {
            console.log('哈哈哈')
        }
        this.hobby = ['篮球'];
    }
    Father.prototype.getType = function () {
        console.log(this.type);
    }

    Son.prototype = new Father();

    // 子类
    function Son(name, age) {
        Father.call(this, name);
        this.age = age;
    }

    var son = new Son('小明', 10);
    son.getType();
</script>
```

常用的继承方式唯一的缺点是，父类的构造函数会被调用两次

#### **(4) 寄生式继承**

```js
<script>
    function madeObj(obj) {
        var newObj = Object.create(obj);
        newObj.name = '小三';
        newObj.age = 3;
        newObj.say = function() {
            console.log('我会说话')
        } 
        return newObj;
    }

    var zhangSan = {
        fatherName: '张三',
        age: 50,
    }
    var xiaoSan = madeObj(zhangSan); 
    console.log(xiaoSan);
</script>
```

就是创建一个封装的函数来增强原有对象的属性，跟借用构造函数一样，每个实例都会创建一遍方法

#### **(5) 寄生组合式继承**

es5最完美的继承方式

```js

```

这是es5最好的继承方式，集合了所有继承方式的优点于一身。

- 原型继承可以访问父类原型上定义的方法
- 寄生继承,父类构造器不需要被调用两次
- 借用构造函数,可以把父类构造器中的属性放入子类的实例



总结: 三种简单的继承方式

1. 原型式继承
2. 借用构造函数继承
3. 寄生式继承

两种复杂的继承方式

1. 组合式继承: 1+2的组合
2. 寄生组合式继承: 1+2+3的组合

 

## 作业

1. 说出以下代码运行结果, 并解释为什么

   ```JavaScript
   var param = 1;
   function main() { 
     console.log('1',param);
     var param = 2;
     console.log('2',this.param);
     this.param = 3;
   }
   
   main(); 
   var m = new main(); 
   ```

2. 说出以下代码运行结果, 并解释为什么

   ```JavaScript
   function Foo() {
     getName = function() {
       console.log('1');
     };
     return this;
   } 
   Foo.getName = function() {
     console.log('2');
   } 
   Foo.prototype.getName = function() {
     console.log('3');
   }
   // 函数表达式
   var getName = function() {
     console.log('4');
   }
   // 覆盖了上面的方法
   function getName() {
     console.log('5');
   }
   
   Foo.getName(); // 调用静态方法 2
   getName(); // 4 函数声明比变量声明提前
    // this=>window, window.getName() 1
    // 要留意getName=xxx, 前面没有var,所以变成了全局函数
   Foo().getName();
   getName();  // 1
   new Foo.getName();  // 2
   new Foo().getName(); // new Foo()得到一个空对象, 这里执行了它原型上的方法 3
   ```

   

